@@ -593,14 +593,8 @@ bsd_new_sta(struct bsd_driver_data *drv, u8 addr[IEEE80211_ADDR_LEN])
 	 */
 	new_assoc = (sta->flags & WLAN_STA_ASSOC) == 0;
 	sta->flags |= WLAN_STA_ASSOC;
-	if (new_assoc) {
-		if (conf->wpa)
-			wpa_sm_event(hapd, sta, WPA_ASSOC);
-		hostapd_new_assoc_sta(hapd, sta);
-	} else {
-		if (conf->wpa)
-			wpa_sm_event(hapd, sta, WPA_REAUTH);
-	}
+	wpa_sm_event(hapd, sta, WPA_ASSOC);
+	hostapd_new_assoc_sta(hapd, sta, !new_assoc);
 	ieee802_1x_notify_port_enabled(sta->eapol_sm, 1);
 	return 0;
 }
@@ -738,7 +732,7 @@ bsd_send_eapol(void *priv, u8 *addr, u8 *data, size_t data_len, int encrypt)
 	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MSGDUMPS))
 		hostapd_hexdump("TX EAPOL", bp, len);
 
-	status = l2_packet_send(drv->sock_xmit, bp, len);
+	status = l2_packet_send(drv->sock_xmit, addr, ETH_P_EAPOL, bp, len);
 
 	if (bp != buf)
 		free(bp);
@@ -746,7 +740,7 @@ bsd_send_eapol(void *priv, u8 *addr, u8 *data, size_t data_len, int encrypt)
 }
 
 static void
-handle_read(void *ctx, unsigned char *src_addr, unsigned char *buf, size_t len)
+handle_read(void *ctx, const u8 *src_addr, const u8 *buf, size_t len)
 {
 	struct bsd_driver_data *drv = ctx;
 	hostapd *hapd = drv->hapd;
@@ -759,7 +753,8 @@ handle_read(void *ctx, unsigned char *src_addr, unsigned char *buf, size_t len)
 		/* XXX cannot happen */
 		return;
 	}
-	ieee802_1x_receive(hapd, src_addr, buf, len);
+	ieee802_1x_receive(hapd, src_addr, buf + sizeof(struct l2_ethhdr),
+			   len - sizeof(struct l2_ethhdr));
 }
 
 static int
@@ -809,7 +804,7 @@ bsd_init(struct hostapd_data *hapd)
 	memcpy(drv->iface, hapd->conf->iface, sizeof(drv->iface));
 
 	drv->sock_xmit = l2_packet_init(drv->iface, NULL, ETH_P_EAPOL,
-				handle_read, drv);
+					handle_read, drv, 1);
 	if (drv->sock_xmit == NULL)
 		goto bad;
 	if (l2_packet_get_own_addr(drv->sock_xmit, hapd->own_addr))
