@@ -32,7 +32,7 @@
 #include <net80211/ieee80211_ioctl.h>
 
 #include <net/if_arp.h>
-#include <linux/wireless.h>
+#include "wireless_copy.h"
 
 #include <netinet/in.h>
 #include <netpacket/packet.h>
@@ -44,6 +44,7 @@
 #include "priv_netlink.h"
 #include "sta_info.h"
 #include "l2_packet.h"
+#include "hostap_common.h"
 
 #include "eapol_sm.h"
 #include "wpa.h"
@@ -63,6 +64,7 @@ struct madwifi_driver_data {
 	struct l2_packet_data *sock_recv;	/* raw packet recv socket */
 	int	ioctl_sock;			/* socket for ioctl() use */
 	int	wext_sock;			/* socket for wireless events */
+	int	we_version;
 	u8	acct_mac[ETH_ALEN];
 	struct hostap_sta_driver_data acct_data;
 };
@@ -730,10 +732,7 @@ madwifi_new_sta(struct madwifi_driver_data *drv, u8 addr[IEEE80211_ADDR_LEN])
 	new_assoc = (sta->flags & WLAN_STA_ASSOC) == 0;
 	sta->flags |= WLAN_STA_ASSOC;
 	wpa_sm_event(hapd, sta, WPA_ASSOC);
-	if (new_assoc)
-		hostapd_new_assoc_sta(hapd, sta);
-	else
-		wpa_sm_event(hapd, sta, WPA_REAUTH);
+	hostapd_new_assoc_sta(hapd, sta, !new_assoc);
 	ieee802_1x_notify_port_enabled(sta->eapol_sm, 1);
 	return 0;
 }
@@ -796,18 +795,35 @@ madwifi_wireless_event_wireless(struct madwifi_driver_data *drv,
 					    char *data, int len)
 {
 	struct hostapd_data *hapd = drv->hapd;
-	struct iw_event *iwe;
+	struct iw_event iwe_buf, *iwe = &iwe_buf;
 	char *pos, *end, *custom, *buf;
 
 	pos = data;
 	end = data + len;
 
 	while (pos + IW_EV_LCP_LEN <= end) {
-		iwe = (struct iw_event *) pos;
+		/* Event data may be unaligned, so make a local, aligned copy
+		 * before processing. */
+		memcpy(&iwe_buf, pos, IW_EV_LCP_LEN);
 		HOSTAPD_DEBUG(HOSTAPD_DEBUG_VERBOSE, "Wireless event: "
 			      "cmd=0x%x len=%d\n", iwe->cmd, iwe->len);
 		if (iwe->len <= IW_EV_LCP_LEN)
 			return;
+
+		custom = pos + IW_EV_POINT_LEN;
+		if (drv->we_version > 18 &&
+		    (iwe->cmd == IWEVMICHAELMICFAILURE ||
+		     iwe->cmd == IWEVCUSTOM)) {
+			/* WE-19 removed the pointer from struct iw_point */
+			char *dpos = (char *) &iwe_buf.u.data.length;
+			int dlen = dpos - (char *) &iwe_buf;
+			memcpy(dpos, pos + IW_EV_LCP_LEN,
+			       sizeof(struct iw_event) - dlen);
+		} else {
+			memcpy(&iwe_buf, pos, sizeof(struct iw_event));
+			custom += IW_EV_POINT_OFF;
+		}
+
 		switch (iwe->cmd) {
 		case IWEVEXPIRED:
 			madwifi_del_sta(drv, iwe->u.addr.sa_data);
@@ -816,7 +832,6 @@ madwifi_wireless_event_wireless(struct madwifi_driver_data *drv,
 			madwifi_new_sta(drv, iwe->u.addr.sa_data);
 			break;
 		case IWEVCUSTOM:
-			custom = pos + IW_EV_POINT_LEN;
 			if (custom + iwe->u.data.length > end)
 				return;
 			buf = malloc(iwe->u.data.length + 1);
@@ -919,6 +934,53 @@ madwifi_wireless_event_receive(int sock, void *eloop_ctx, void *sock_ctx)
 }
 
 
+static int
+madwifi_get_we_version(struct madwifi_driver_data *drv)
+{
+	struct iw_range *range;
+	struct iwreq iwr;
+	int minlen;
+	size_t buflen;
+
+	drv->we_version = 0;
+
+	/*
+	 * Use larger buffer than struct iw_range in order to allow the
+	 * structure to grow in the future.
+	 */
+	buflen = sizeof(struct iw_range) + 500;
+	range = malloc(buflen);
+	if (range == NULL)
+		return -1;
+	memset(range, 0, buflen);
+
+	memset(&iwr, 0, sizeof(iwr));
+	strncpy(iwr.ifr_name, drv->iface, IFNAMSIZ);
+	iwr.u.data.pointer = (caddr_t) range;
+	iwr.u.data.length = buflen;
+
+	minlen = ((char *) &range->enc_capa) - (char *) range +
+		sizeof(range->enc_capa);
+
+	if (ioctl(drv->ioctl_sock, SIOCGIWRANGE, &iwr) < 0) {
+		perror("ioctl[SIOCGIWRANGE]");
+		free(range);
+		return -1;
+	} else if (iwr.u.data.length >= minlen &&
+		   range->we_version_compiled >= 18) {
+		wpa_printf(MSG_DEBUG, "SIOCGIWRANGE: WE(compiled)=%d "
+			   "WE(source)=%d enc_capa=0x%x",
+			   range->we_version_compiled,
+			   range->we_version_source,
+			   range->enc_capa);
+		drv->we_version = range->we_version_compiled;
+	}
+
+	free(range);
+	return 0;
+}
+
+
 static int
 madwifi_wireless_event_init(void *priv)
 {
@@ -926,6 +988,8 @@ madwifi_wireless_event_init(void *priv)
 	int s;
 	struct sockaddr_nl local;
 
+	madwifi_get_we_version(drv);
+
 	drv->wext_sock = -1;
 
 	s = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
@@ -999,7 +1063,7 @@ madwifi_send_eapol(void *priv, u8 *addr, u8 *data, size_t data_len, int encrypt)
 	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MSGDUMPS))
 		hostapd_hexdump("TX EAPOL", bp, len);
 
-	status = l2_packet_send(drv->sock_xmit, bp, len);
+	status = l2_packet_send(drv->sock_xmit, addr, ETH_P_EAPOL, bp, len);
 
 	if (bp != buf)
 		free(bp);
@@ -1007,7 +1071,7 @@ madwifi_send_eapol(void *priv, u8 *addr, u8 *data, size_t data_len, int encrypt)
 }
 
 static void
-handle_read(void *ctx, unsigned char *src_addr, unsigned char *buf, size_t len)
+handle_read(void *ctx, const u8 *src_addr, const u8 *buf, size_t len)
 {
 	struct madwifi_driver_data *drv = ctx;
 	hostapd *hapd = drv->hapd;
@@ -1020,7 +1084,8 @@ handle_read(void *ctx, unsigned char *src_addr, unsigned char *buf, size_t len)
 		/* XXX cannot happen */
 		return;
 	}
-	ieee802_1x_receive(hapd, src_addr, buf, len);
+	ieee802_1x_receive(hapd, src_addr, buf + sizeof(struct l2_ethhdr),
+			   len - sizeof(struct l2_ethhdr));
 }
 
 static int
@@ -1055,7 +1120,7 @@ madwifi_init(struct hostapd_data *hapd)
 	drv->ifindex = ifr.ifr_ifindex;
 
 	drv->sock_xmit = l2_packet_init(drv->iface, NULL, ETH_P_EAPOL,
-				handle_read, drv);
+					handle_read, drv, 1);
 	if (drv->sock_xmit == NULL)
 		goto bad;
 	if (l2_packet_get_own_addr(drv->sock_xmit, hapd->own_addr))
@@ -1065,7 +1130,8 @@ madwifi_init(struct hostapd_data *hapd)
 			"Configure bridge %s for EAPOL traffic.\n",
 			hapd->conf->bridge);
 		drv->sock_recv = l2_packet_init(hapd->conf->bridge, NULL,
-						ETH_P_EAPOL, handle_read, drv);
+						ETH_P_EAPOL, handle_read, drv,
+						0);
 		if (drv->sock_recv == NULL)
 			goto bad;
 	} else
