@@ -806,10 +806,7 @@ static void prism54_handle_assoc(struct prism54_driver_data *drv,
 		new_assoc = (sta->flags & WLAN_STA_ASSOC) == 0;
 		sta->flags |= WLAN_STA_ASSOC;
 		wpa_sm_event(drv->hapd, sta, WPA_ASSOC);
-		if (new_assoc)
-			hostapd_new_assoc_sta(drv->hapd, sta);
-		else
-			wpa_sm_event(drv->hapd, sta, WPA_REAUTH);
+		hostapd_new_assoc_sta(drv->hapd, sta, !new_assoc);
 		ieee802_1x_notify_port_enabled(sta->eapol_sm, 1);
 		sta->timeout_next = STA_NULLFUNC;
 		return;
