@@ -54,6 +54,9 @@ extern const struct eap_method eap_method_sim;
 #ifdef EAP_PAX
 extern const struct eap_method eap_method_pax;
 #endif /* EAP_PAX */
+#ifdef EAP_PSK
+extern const struct eap_method eap_method_psk;
+#endif /* EAP_PSK */
 
 static const struct eap_method *eap_methods[] =
 {
@@ -85,6 +88,9 @@ static const struct eap_method *eap_methods[] =
 #ifdef EAP_PAX
 	&eap_method_pax,
 #endif /* EAP_PAX */
+#ifdef EAP_PSK
+	&eap_method_psk,
+#endif /* EAP_PSK */
 };
 #define NUM_EAP_METHODS (sizeof(eap_methods) / sizeof(eap_methods[0]))
 
