@@ -66,6 +66,9 @@ static void sm_ ## machine ## _Step(struct eapol_state_machine *sm)
 #define SM_STEP_RUN(machine) sm_ ## machine ## _Step(sm)
 
 
+static void eapol_sm_step_run(struct eapol_state_machine *sm);
+static void eapol_sm_step_cb(void *eloop_ctx, void *timeout_ctx);
+
 
 /* Port Timers state machine - implemented as a function that will be called
  * once a second as a registered event loop timeout */
@@ -74,19 +77,34 @@ static void eapol_port_timers_tick(void *eloop_ctx, void *timeout_ctx)
 {
 	struct eapol_state_machine *state = timeout_ctx;
 
-	if (state->aWhile > 0)
+	if (state->aWhile > 0) {
 		state->aWhile--;
-	if (state->quietWhile > 0)
+		if (state->aWhile == 0) {
+			wpa_printf(MSG_DEBUG, "IEEE 802.1X: " MACSTR
+				   " - aWhile --> 0",
+				   MAC2STR(state->addr));
+		}
+	}
+
+	if (state->quietWhile > 0) {
 		state->quietWhile--;
-	if (state->reAuthWhen > 0)
+		if (state->quietWhile == 0) {
+			wpa_printf(MSG_DEBUG, "IEEE 802.1X: " MACSTR
+				   " - quietWhile --> 0",
+				   MAC2STR(state->addr));
+		}
+	}
+
+	if (state->reAuthWhen > 0) {
 		state->reAuthWhen--;
+		if (state->reAuthWhen == 0) {
+			wpa_printf(MSG_DEBUG, "IEEE 802.1X: " MACSTR
+				   " - reAuthWhen --> 0",
+				   MAC2STR(state->addr));
+		}
+	}
 
-	if (state->hapd->conf->debug >= HOSTAPD_DEBUG_MSGDUMPS)
-		printf("IEEE 802.1X: " MACSTR " Port Timers TICK "
-		       "(timers: %d %d %d)\n", MAC2STR(state->addr),
-		       state->aWhile, state->quietWhile, state->reAuthWhen);
-
-	eapol_sm_step(state);
+	eapol_sm_step_run(state);
 
 	eloop_register_timeout(1, 0, eapol_port_timers_tick, eloop_ctx, state);
 }
@@ -727,6 +745,7 @@ void eapol_sm_free(struct eapol_state_machine *sm)
 		return;
 
 	eloop_cancel_timeout(eapol_port_timers_tick, sm->hapd, sm);
+	eloop_cancel_timeout(eapol_sm_step_cb, sm, NULL);
 	if (sm->eap)
 		eap_sm_deinit(sm->eap);
 	free(sm);
@@ -743,55 +762,51 @@ static int eapol_sm_sta_entry_alive(struct hostapd_data *hapd, u8 *addr)
 }
 
 
-void eapol_sm_step(struct eapol_state_machine *sm)
+static void eapol_sm_step_run(struct eapol_state_machine *sm)
 {
 	struct hostapd_data *hapd = sm->hapd;
 	u8 addr[ETH_ALEN];
 	int prev_auth_pae, prev_be_auth, prev_reauth_timer, prev_auth_key_tx,
 		prev_key_rx, prev_ctrl_dir;
 
-	/* FIX: could re-run eapol_sm_step from registered timeout (after
-	 * 0 sec) to make sure that other possible timeouts/events are
-	 * processed */
-
 	memcpy(addr, sm->sta->addr, ETH_ALEN);
-restart:
-	do {
-		prev_auth_pae = sm->auth_pae.state;
-		prev_be_auth = sm->be_auth.state;
-		prev_reauth_timer = sm->reauth_timer.state;
-		prev_auth_key_tx = sm->auth_key_tx.state;
-		prev_key_rx = sm->key_rx.state;
-		prev_ctrl_dir = sm->ctrl_dir.state;
 
-		SM_STEP_RUN(AUTH_PAE);
-		if (!sm->initializing && !eapol_sm_sta_entry_alive(hapd, addr))
-			break;
+	prev_auth_pae = sm->auth_pae.state;
+	prev_be_auth = sm->be_auth.state;
+	prev_reauth_timer = sm->reauth_timer.state;
+	prev_auth_key_tx = sm->auth_key_tx.state;
+	prev_key_rx = sm->key_rx.state;
+	prev_ctrl_dir = sm->ctrl_dir.state;
+
+	SM_STEP_RUN(AUTH_PAE);
+	if (sm->initializing || eapol_sm_sta_entry_alive(hapd, addr))
 		SM_STEP_RUN(BE_AUTH);
-		if (!sm->initializing && !eapol_sm_sta_entry_alive(hapd, addr))
-			break;
+	if (sm->initializing || eapol_sm_sta_entry_alive(hapd, addr))
 		SM_STEP_RUN(REAUTH_TIMER);
-		if (!sm->initializing && !eapol_sm_sta_entry_alive(hapd, addr))
-			break;
+	if (sm->initializing || eapol_sm_sta_entry_alive(hapd, addr))
 		SM_STEP_RUN(AUTH_KEY_TX);
-		if (!sm->initializing && !eapol_sm_sta_entry_alive(hapd, addr))
-			break;
+	if (sm->initializing || eapol_sm_sta_entry_alive(hapd, addr))
 		SM_STEP_RUN(KEY_RX);
-		if (!sm->initializing && !eapol_sm_sta_entry_alive(hapd, addr))
-			break;
+	if (sm->initializing || eapol_sm_sta_entry_alive(hapd, addr))
 		SM_STEP_RUN(CTRL_DIR);
-		if (!sm->initializing && !eapol_sm_sta_entry_alive(hapd, addr))
-			break;
-	} while (prev_auth_pae != sm->auth_pae.state ||
-		 prev_be_auth != sm->be_auth.state ||
-		 prev_reauth_timer != sm->reauth_timer.state ||
-		 prev_auth_key_tx != sm->auth_key_tx.state ||
-		 prev_key_rx != sm->key_rx.state ||
-		 prev_ctrl_dir != sm->ctrl_dir.state);
+
+	if (prev_auth_pae != sm->auth_pae.state ||
+	    prev_be_auth != sm->be_auth.state ||
+	    prev_reauth_timer != sm->reauth_timer.state ||
+	    prev_auth_key_tx != sm->auth_key_tx.state ||
+	    prev_key_rx != sm->key_rx.state ||
+	    prev_ctrl_dir != sm->ctrl_dir.state) {
+		/* Re-run from eloop timeout */
+		eapol_sm_step(sm);
+		return;
+	}
 
 	if (eapol_sm_sta_entry_alive(hapd, addr) && sm->eap) {
-		if (eap_sm_step(sm->eap))
-			goto restart;
+		if (eap_sm_step(sm->eap)) {
+			/* Re-run from eloop timeout */
+			eapol_sm_step(sm);
+			return;
+		}
 	}
 
 	if (eapol_sm_sta_entry_alive(hapd, addr))
@@ -799,15 +814,34 @@ restart:
 }
 
 
+static void eapol_sm_step_cb(void *eloop_ctx, void *timeout_ctx)
+{
+	struct eapol_state_machine *sm = eloop_ctx;
+	eapol_sm_step_run(sm);
+}
+
+
+void eapol_sm_step(struct eapol_state_machine *sm)
+{
+	/*
+	 * Run eapol_sm_step_run from a registered timeout to make sure that
+	 * other possible timeouts/events are processed and to avoid long
+	 * function call chains.
+	 */
+
+	eloop_register_timeout(0, 0, eapol_sm_step_cb, sm, NULL);
+}
+
+
 void eapol_sm_initialize(struct eapol_state_machine *sm)
 {
 	sm->initializing = TRUE;
 	/* Initialize the state machines by asserting initialize and then
 	 * deasserting it after one step */
 	sm->initialize = TRUE;
-	eapol_sm_step(sm);
+	eapol_sm_step_run(sm);
 	sm->initialize = FALSE;
-	eapol_sm_step(sm);
+	eapol_sm_step_run(sm);
 	sm->initializing = FALSE;
 
 	/* Start one second tick for port timers state machine */
