@@ -126,6 +126,7 @@ void hostapd_logger(struct hostapd_data *hapd, const u8 *addr,
 			 module_str, module_str ? ": " : "", fmt);
 
 	if ((conf_stdout & module) && level >= conf_stdout_level) {
+		wpa_debug_print_timestamp();
 		vprintf(format, ap);
 		printf("\n");
 	}
@@ -198,20 +199,24 @@ static void hostapd_deauth_all_stas(hostapd *hapd)
 
 
 /* This function will be called whenever a station associates with the AP */
-void hostapd_new_assoc_sta(hostapd *hapd, struct sta_info *sta)
+void hostapd_new_assoc_sta(hostapd *hapd, struct sta_info *sta, int reassoc)
 {
 	/* IEEE 802.11F (IAPP) */
 	if (hapd->conf->ieee802_11f)
 		iapp_new_station(hapd->iapp, sta);
 
-	/* Start accounting here, if IEEE 802.1X is not used. IEEE 802.1X code
-	 * will start accounting after the station has been authorized. */
-	if (!hapd->conf->ieee802_1x)
+	/* Start accounting here, if IEEE 802.1X and WPA are not used.
+	 * IEEE 802.1X/WPA code will start accounting after the station has
+	 * been authorized. */
+	if (!hapd->conf->ieee802_1x && !hapd->conf->wpa)
 		accounting_sta_start(hapd, sta);
 
 	/* Start IEEE 802.1X authentication process for new stations */
 	ieee802_1x_new_station(hapd, sta);
-	wpa_new_station(hapd, sta);
+	if (reassoc)
+		wpa_sm_event(hapd, sta, WPA_REAUTH);
+	else
+		wpa_new_station(hapd, sta);
 }
 
 
