@@ -18,6 +18,7 @@
 #include <string.h>
 #include <sys/socket.h>
 #include <net/if.h>
+#include <netinet/in.h>
 #include <arpa/inet.h>
 #include <time.h>
 
@@ -34,6 +35,7 @@
 #include "wpa.h"
 #include "accounting.h"
 #include "driver.h"
+#include "hostap_common.h"
 
 
 static u8 * hostapd_eid_supp_rates(hostapd *hapd, u8 *eid)
@@ -1093,10 +1095,7 @@ static void handle_assoc_cb(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	}
 
 	wpa_sm_event(hapd, sta, WPA_ASSOC);
-	if (new_assoc)
-		hostapd_new_assoc_sta(hapd, sta);
-	else
-		wpa_sm_event(hapd, sta, WPA_REAUTH);
+	hostapd_new_assoc_sta(hapd, sta, !new_assoc);
 
 	ieee802_1x_notify_port_enabled(sta->eapol_sm, 1);
 
