@@ -17,6 +17,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
+#include <netinet/in.h>
 #include <arpa/inet.h>
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -27,6 +28,7 @@
 #include "radius.h"
 #include "radius_client.h"
 #include "eloop.h"
+#include "hostap_common.h"
 
 #define RADIUS_ACL_TIMEOUT 30
 
