@@ -670,7 +670,8 @@ static void handle_eap(hostapd *hapd, struct sta_info *sta, u8 *buf,
 
 
 /* Process the EAPOL frames from the Supplicant */
-void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len)
+void ieee802_1x_receive(struct hostapd_data *hapd, const u8 *sa, const u8 *buf,
+			size_t len)
 {
 	struct sta_info *sta;
 	struct ieee802_1x_hdr *hdr;
@@ -805,20 +806,17 @@ void ieee802_1x_new_station(hostapd *hapd, struct sta_info *sta)
 	if (!hapd->conf->ieee802_1x || sta->wpa_key_mgmt == WPA_KEY_MGMT_PSK)
 		return;
 
-	if (sta->eapol_sm) {
-		sta->eapol_sm->portEnabled = TRUE;
-		eapol_sm_step(sta->eapol_sm);
-		return;
-	}
-
-	hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_IEEE8021X,
-		       HOSTAPD_LEVEL_DEBUG, "start authentication");
-	sta->eapol_sm = eapol_sm_alloc(hapd, sta);
 	if (sta->eapol_sm == NULL) {
 		hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_IEEE8021X,
-			       HOSTAPD_LEVEL_INFO, "failed to allocate "
-			       "state machine");
-		return;
+			       HOSTAPD_LEVEL_DEBUG, "start authentication");
+		sta->eapol_sm = eapol_sm_alloc(hapd, sta);
+		if (sta->eapol_sm == NULL) {
+			hostapd_logger(hapd, sta->addr,
+				       HOSTAPD_MODULE_IEEE8021X,
+				       HOSTAPD_LEVEL_INFO,
+				       "failed to allocate state machine");
+			return;
+		}
 	}
 
 	sta->eapol_sm->portEnabled = TRUE;
@@ -836,20 +834,49 @@ void ieee802_1x_new_station(hostapd *hapd, struct sta_info *sta)
 		sta->eapol_sm->authSuccess = TRUE;
 		if (sta->eapol_sm->eap)
 			eap_sm_notify_cached(sta->eapol_sm->eap);
-	}
+	} else
+		eapol_sm_step(sta->eapol_sm);
 }
 
 
-static void ieee802_1x_free_radius_class(struct eapol_state_machine *sm)
+void ieee802_1x_free_radius_class(struct radius_class_data *class)
 {
 	int i;
-	if (sm->radius_class == NULL)
+	if (class == NULL)
 		return;
-	for (i = 0; i < sm->radius_class_count; i++)
-		free(sm->radius_class[i].data);
-	free(sm->radius_class);
-	sm->radius_class = NULL;
-	sm->radius_class_count = 0;
+	for (i = 0; i < class->count; i++)
+		free(class->attr[i].data);
+	free(class->attr);
+	class->attr = NULL;
+	class->count = 0;
+}
+
+
+int ieee802_1x_copy_radius_class(struct radius_class_data *dst,
+				 struct radius_class_data *src)
+{
+	size_t i;
+
+	if (src->attr == NULL)
+		return 0;
+
+	dst->attr = malloc(src->count * sizeof(struct radius_attr_data));
+	if (dst->attr == NULL)
+		return -1;
+
+	memset(dst->attr, 0, src->count * sizeof(struct radius_attr_data));
+	dst->count = 0;
+
+	for (i = 0; i < src->count; i++) {
+		dst->attr[i].data = malloc(src->attr[i].len);
+		if (dst->attr[i].data == NULL)
+			break;
+		dst->count++;
+		memcpy(dst->attr[i].data, src->attr[i].data, src->attr[i].len);
+		dst->attr[i].len = src->attr[i].len;
+	}
+
+	return 0;
 }
 
 
@@ -872,7 +899,7 @@ void ieee802_1x_free_station(struct sta_info *sta)
 	free(sm->last_eap_supp);
 	free(sm->last_eap_radius);
 	free(sm->identity);
-	ieee802_1x_free_radius_class(sm);
+	ieee802_1x_free_radius_class(&sm->radius_class);
 	free(sm->eapol_key_sign);
 	free(sm->eapol_key_crypt);
 	eapol_sm_free(sm);
@@ -1023,7 +1050,7 @@ static void ieee802_1x_store_radius_class(struct hostapd_data *hapd,
 	    sm == NULL)
 		return;
 
-	ieee802_1x_free_radius_class(sm);
+	ieee802_1x_free_radius_class(&sm->radius_class);
 	count = radius_msg_count_attr(msg, RADIUS_ATTR_CLASS, 1);
 	if (count <= 0)
 		return;
@@ -1055,11 +1082,12 @@ static void ieee802_1x_store_radius_class(struct hostapd_data *hapd,
 		nclass_count++;
 	}
 
-	sm->radius_class = nclass;
-	sm->radius_class_count = nclass_count;
-	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "IEEE 802.1X: Stored %d RADIUS "
+	sm->radius_class.attr = nclass;
+	sm->radius_class.count = nclass_count;
+	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "IEEE 802.1X: Stored %lu RADIUS "
 		      "Class attributes for " MACSTR "\n",
-		      sm->radius_class_count, MAC2STR(sta->addr));
+		      (unsigned long) sm->radius_class.count,
+		      MAC2STR(sta->addr));
 }
 
 
@@ -1190,12 +1218,12 @@ ieee802_1x_receive_auth(struct radius_msg *msg, struct radius_msg *req,
 		override_eapReq = 1;
 		ieee802_1x_get_keys(hapd, sta, msg, req, shared_secret,
 				    shared_secret_len);
+		ieee802_1x_store_radius_class(hapd, sta, msg);
 		if (sm->keyAvailable) {
 			pmksa_cache_add(hapd, sta, sm->eapol_key_crypt,
 					session_timeout_set ?
 					session_timeout : -1);
 		}
-		ieee802_1x_store_radius_class(hapd, sta, msg);
 		break;
 	case RADIUS_CODE_ACCESS_REJECT:
 		sm->eapFail = TRUE;
@@ -1520,12 +1548,12 @@ u8 * ieee802_1x_get_identity(struct eapol_state_machine *sm, size_t *len)
 u8 * ieee802_1x_get_radius_class(struct eapol_state_machine *sm, size_t *len,
 				 int idx)
 {
-	if (sm == NULL || sm->radius_class == NULL ||
-	    idx >= sm->radius_class_count)
+	if (sm == NULL || sm->radius_class.attr == NULL ||
+	    idx >= sm->radius_class.count)
 		return NULL;
 
-	*len = sm->radius_class[idx].len;
-	return sm->radius_class[idx].data;
+	*len = sm->radius_class.attr[idx].len;
+	return sm->radius_class.attr[idx].data;
 }
 
 
@@ -1545,6 +1573,7 @@ void ieee802_1x_notify_port_enabled(struct eapol_state_machine *sm,
 	if (sm == NULL)
 		return;
 	sm->portEnabled = enabled ? TRUE : FALSE;
+	eapol_sm_step(sm);
 }
 
 
@@ -1554,6 +1583,7 @@ void ieee802_1x_notify_port_valid(struct eapol_state_machine *sm,
 	if (sm == NULL)
 		return;
 	sm->portValid = valid ? TRUE : FALSE;
+	eapol_sm_step(sm);
 }
 
 
